

#if defined(OS_WIN32)
#include <windows.h>
#include <malloc.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <math.h>


#include "pres2prot.h"
#include "pres2protfuncs.h"
#include "colortools.h"
#include "BVQXPluginInterface.h"
#include "Plugin_PRT_Header.h"
#include "PRES_Header.h"



ProtocolGenerator::ProtocolGenerator()
{
}


ProtocolGenerator::~ProtocolGenerator()
{
}


using namespace std;

enum menuoptions {MENU_CREATEPRT = 1, MENU_OFFSETPRT, MENU_SORTPRT, MENU_SORTAUTOCOLOR};


char descr[] = "~~~ protocolgenerator (v0.9) ~~~"
	"<BR>Generates a BrainVoyager QX Protocol file (*.prt) in milliseconds from a Presentation Log file (*.log)."
	"<BR>Each Picture, Sound, Video or Response event will result in a new condition."
	"<BR>It is also possible to shift the offset of all intervals by a certain amount (in milliseconds)."
        "<BR>Other functions are sorting the conditions of an existing protocol."
	"<BR>Version: 0.9.1";
static const char *pluginDescription = descr;

bool ProtocolGenerator::initPlugin() // this function is called in GUI plugins once after class has been created - not called in pure computational plugins
{
	// This function call and check of return value is necessary to check whether access to QX works
	int ret = InitQXAccess();
	
	// Never continue, if the "InitQXAccess" function fails
	if(!ret) {
		qxLogText("init failed.");
		return false;
	}
	
	// Test whether the used QX version is new enough to support the commands used in your plugin
	int Major, Minor, SubMinor;
	qxGetVersionOfBrainVoyagerQX(&Major, &Minor, &SubMinor);
	if( !checkVersionGreaterEqualVersion(Major, Minor, SubMinor, 2, 2, 0) ) {
		// we can use message box since this was available with the first QX version with plugin support (v1.3.8)
		qxMessageBox("Can not execute this plugin - BrainVoyager QX version 2.2 or higher required.");
		return false;
	}
	return true;
}

bool ProtocolGenerator::execute() {

	bool cancel = false;
	bool success = true;
	qxShowLogTab();
	qxLogText("\nprotocolgenerator> Started...");

	int usrResp;
	char *instruction		= "Please select the Presentation Log file";
	char *defaultFolder		= new char[256];
	qxGetBrainVoyagerQXPath(defaultFolder);
	char *typeOfFiles		= "Presentation log file (*.log)";
	char *filename			= new char[256];

	char *intro = new char[1024];
	sprintf(intro, "Please select an option:<P>1 = Create stimulation protocol from Presentation Log file");
	sprintf(intro + strlen(intro), "<BR>2 = Change the offset in an existing stimulation protocol");
	sprintf(intro + strlen(intro), "<BR>3 = Sort conditions in an existing stimulation protocol");
    //    sprintf(intro + strlen(intro), "<BR>4 = Sort and autocolor conditions of existing protocol (Sanae only)");
        int pluginAction = qxGetIntegerInput(intro, 1,1,3);

	if (pluginAction == MENU_CREATEPRT) {

			qxMessageBox("This function generates a BrainVoyager QX stimulation protocol file (*.prt),\nbased on a Presentation Log file (*.log)");
			int DimX, DimY, DimZ;
			unsigned char *VMR = qxGetCurrentVMR(&DimX, &DimY, &DimZ);
			if(VMR == 0) {
				qxMessageBox("If no project is opened in BrainVoyager, the conditions cannot be added to the protocol.");
				qxLogText( "No VMR present, action canceled.");
				return false;
			}
			usrResp = qxGetOpenFileName(instruction, "", typeOfFiles, filename);
			if (usrResp == 1) {

				char columnreprinfo[] = "Which fields are in the log file?<P>"
				"1 = Trial, EventType, Code, Time, TTime, Uncertainty, Duration<BR>"
				"2 = Subject, Trial, EventType, Code, Time, TTime, Uncertainty, Duration<BR>";
				int fieldrepres = qxGetIntegerInput(columnreprinfo, 1, 1,2);
                bool pictureEventsOnly = (bool) qxYesNoMessageBox("Select 'Picture' events only?<BR>(Otherwise, Response events can result in '1', '2', '3' etc. condition names of button box numbers)");
				bool readsubj = (fieldrepres == 2);
				qxLogText("Read Presentation log file...");
				presHdr *pres = readpres(filename, readsubj, pictureEventsOnly);
				qxLogText("Converting to BrainVoyager stimulation protocol...");
				int defaultTR = 2000;
				int usrTR = qxGetIntegerInput("What is the TR in milliseconds?", defaultTR, 1,100000);
	//			int inMsec = qxYesNoMessageBox("Would you prefer the protocol in milliseconds (yes) or volumes (no)?");
				int resolution = UNITS_MSEC;// ((inMsec == 1) ? UNITS_MSEC : UNITS_VOLS);
				int margin = 0;
				if (resolution == UNITS_VOLS) {
					margin = qxGetIntegerInput("If necessary, enter a margin (msec) to decide to which condition a volume should be assigned", 0, 0, usrTR);
				}
				int usrDur = qxGetIntegerInput("Please select the computation of event duration:\n\n1 = Based on time until next event (Picture, Sound, Video, Response)\n2 = Based on Duration parameter in Presentation log file", 1, 1,2);
				if (!(usrDur == DURATION_LOG_BASED || usrDur == DURATION_NEXT_PIC_BASED)) {
					success = false;
				}
				if (success) {
					success = convertprotocol(pres, usrTR, resolution, usrDur);
				}

			} else {
				cancel = true;
			}

	} else if (pluginAction == MENU_OFFSETPRT) {

		Protocol_Header *prt = new Protocol_Header;
		qxMessageBox("This function changes the offsets from all intervals");
		qxGetProtocol(prt);
		int returnval = qxGetProtocol(prt);
        if (returnval > 0) {
			char *offsetmsg = new char[512];
			if (prt->TimeResolution == UNITS_MSEC) {
			   sprintf(offsetmsg, "Please enter the amount of milliseconds that the intervals should shift");
			} else if (prt->TimeResolution == UNITS_VOLS) {
				sprintf(offsetmsg, "Please enter the number of volumes that the intervals should shift");
			}
			int offset = qxGetIntegerInput(offsetmsg, 0,0,3000000);
			if (offset < 0) {
				success = false;
				cancel = true;
			} else {
				int doSubtract = qxYesNoMessageBox("Should the offset be subtracted (yes) or added (no) to the current intervals?");
				int calculationtype = (doSubtract ? CALC_SUBTRACT : CALC_ADD);
				success = changeProtocolOffset(prt, calculationtype, offset);
			}
		} else {
			qxLogText("No protocol linked.");
			success = false;
			cancel = true;
		}

	} else if (pluginAction == MENU_SORTPRT) {

                Protocol_Header *prt = new Protocol_Header;
                int returnval = qxGetProtocol(prt);
                if (returnval == 1) {

                        vector<int> conditionvector(prt->NrOfConditions);
                  //    char *neworderstr = getNewOrderOfConditions(prt);
						char *return_txt = new char[256];
						int usrreply = getNewOrderOfConditions(prt, return_txt);
						if (usrreply > 0){
//	                        success = convertConditionListToConditionVector(neworderstr, conditionvector);
							success = convertConditionListToConditionVector(return_txt, conditionvector);
		                    success = makeAndSaveReorderedProtocol(prt, conditionvector);
						} else {
							qxLogText("User canceled protocol sorting.");
							success = false;						
						}
                } else {
                        qxLogText("Protocol header could not be retrieved.");
                        success = false;
                }


     } else if (pluginAction == MENU_SORTAUTOCOLOR){

                Protocol_Header *prt = new Protocol_Header;
                int returnval = qxGetProtocol(prt);
                if (returnval == 1) {

                        vector<int> conditionvector(prt->NrOfConditions);
                        int typeOfAnalysis = qxGetIntegerInput("What is the type of analysis?\n\n1 = GLM (14 pred.)\n2 = GLM 5x2 (18 pred.)\n3 = GLM match (16 pred.)\n4 = GLM offerA+playertype (23 pred.)\n5 = GLM exp+playertype (20 pred.)\n6= GLM earn+playertype (20 pred.)\n7= GLM expExp hpc (15 pred.)\n", 1, 1, 7);
                        if (typeOfAnalysis > 0) {
                            char *neworderstr = new char[256];
                            if (typeOfAnalysis == ANALYSIS_GLM) {
                                sprintf(neworderstr, "0,1,13,2,3,4,5,6,7,8,9,10,11,12");
                            } else if (typeOfAnalysis == ANALYSIS_GLM_5X2) {
                                sprintf(neworderstr, "0,1,13,2,3,4,5,6,14,15,16,17,7,8,9,10,11,12");
                            } else if (typeOfAnalysis == ANALYSIS_GLM_MATCH_EARNINGS) {
                                sprintf(neworderstr, "0,1,13,2,3,4,5,6,7,8,9,10,11,14,15,12");
                            } else if (typeOfAnalysis == ANALYSIS_GLM_OFFERA_PLAYERTYPE) {
                                sprintf(neworderstr, "0,1,13,2,3,4,5,14,15,16,17,18,19,20,21,22,6,7,8,9,10,11,12"); // nog testen								
                            } else if (typeOfAnalysis == ANALYSIS_GLM_MATCH_EXP_PLAYERTYPE) {
                                sprintf(neworderstr, "0,1,13,2,3,4,5,6,7,8,9,10,11,14,15,12"); // nog aanpassen
                            } else if (typeOfAnalysis == ANALYSIS_GLM_MATCH_EARNINGS_PLAYERTYPE) {
                                sprintf(neworderstr, "0,1,13,2,3,4,5,6,7,8,9,10,11,14,15,12"); // nog aanpassen
                            } else if (typeOfAnalysis == ANALYSIS_GLM_EXPEXP_HPC) {
								sprintf(neworderstr, "0,1,13,2,14,3,4,5,6,7,8,9,10,11,12"); // added 13-06-11
							}
							success = convertConditionListToConditionVector(neworderstr, conditionvector);
                            success = makeAndSaveReorderedAndColoredProtocol(prt, conditionvector, typeOfAnalysis);
                        }
                } else {
                        qxLogText("Protocol header could not be retrieved.");
                        success = false;
                }

	} else {
		qxLogText("No valid option selected.");
		cancel = true;
	}


	if (cancel) {
		qxLogText("Cancelled further operations.");
	} else if (success) {
		qxLogText("Plugin operations finished successfully.");
	}
	qxLogText("Finished.\n");
	return true;
}


PLUGIN_ACCESS BVQXPluginInterface *createPlugin()
{
	return new ProtocolGenerator();
}

PLUGIN_ACCESS const char *getPluginName()
{
	return "protocolgenerator (v0.9)";
}


PLUGIN_ACCESS const char *getPluginDescription()
{
	return pluginDescription;
}

PLUGIN_ACCESS const char *getAuthor()
{
	return "<i>Hester Breman, Brain Innovation B.V., 2008-2012</i>";
}


PLUGIN_ACCESS const char *getHelpFile()
{
	return "";
}

// you may provide a version string ("Major.Minor") for (updated) releases of your plugins
PLUGIN_ACCESS const char *getPluginVersion()
{
	return "v0.9";
}

// NEW: If you want to use GUI dialogs, return a GUI script file, if not, return empty C string
PLUGIN_ACCESS const char *getGUIScript()
{
	return "";
}



