
#if defined(OS_WIN32)
#include <windows.h>
#include <malloc.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>
#include <string.h>
#include <string> // c++ strings
#include <map> // map
#include <vector>
#include <iterator> // iterate over items in map

#include "BVQXPluginInterface.h"
#include "Plugin_PRT_Header.h"
#include "PRES_Header.h"
#include "pres2protfuncs.h"
#include "colortools.h"
#include "iofuncs.h"

using namespace std;

presHdr *readpres(char *presFileName, bool readsubjectcolumn,  bool pictureEventsOnly) {

	presHdr *pres = new presHdr;
	FILE *presFile;

	try {

		presFile = fopen(presFileName, "r");

		int nroflines = 0;
		int nroftrials = 0;
		int Trial;
		char *Subject = new char[100];
		char *EventType = new char[100];
		char *Code = new char[100];
		double Time;
		int TTime;
		int Uncertainty;
		double Duration;
		char str[256];
		char tmp3[512];
		char eventPicture[] = "Picture";
		char eventSound[] = "Sound";
		char eventVideo[] = "Video";
		char eventResponse[] = "Response";
		char prevline[256];
		bool endoffile = false;
		int nrofconditions = 0;

		char *tmpreport = new char[256];
		int maxlines = 100000;

		while (!endoffile) {

			fgets(str, 256, presFile);
			if (strcmp(str, prevline) == 0) { // if line is empty, last line data are same as for empty line
				endoffile = true;
				maxlines = nroflines;
				break;
			}
			if (nroflines > 5) {
				if (readsubjectcolumn) {
					sscanf(str, "%s %d %s %s %lf %d %d %lf", Subject, &Trial, EventType, Code, &Time, &TTime, &Uncertainty, &Duration); // new in v05
				} else {
					sscanf(str, "%d %s %s %lf %d %d %lf", &Trial, EventType, Code, &Time, &TTime, &Uncertainty, &Duration);
				}
				sprintf(tmpreport, "Trial: %d EventType: %s Code: %s", Trial, EventType, Code);
				qxLogText(tmpreport);
                                if (pictureEventsOnly) {
                                    if (strcmp(EventType, eventPicture) == 0) {
                                        nrofconditions++;
                                    }
                                } else if ((strcmp(EventType, eventPicture) == 0) || (strcmp(EventType, eventSound) == 0)
					|| (strcmp(EventType, eventVideo) == 0) || (strcmp(EventType, eventResponse) == 0)) {
					nrofconditions++;
                                 } // end pictureEventsOnly
                                nroftrials = Trial;
                        } // end readsubjectcolumn
                        strcpy(prevline, str);
			nroflines++;
		}

		sprintf(tmp3, "Number of trials: %d, Picture/Sound/Video/Response events: %d", nroftrials, nrofconditions);
		qxLogText(tmp3);

		sprintf(tmp3, "Number of lines in first part of file: %d", nroflines);
		qxLogText(tmp3);

		presTrial *trials = new presTrial[nrofconditions];

		endoffile = false;
		long int fid = 0L;
		int currentcondition = 0;
		nroflines = 0;
		fseek(presFile, fid, SEEK_SET);
		while (!endoffile) {

			fgets(str, 256, presFile);
			if (strcmp(str, prevline) == 0) {
				endoffile = true;
				break;
			}
			if ((nroflines > 5) && (nroflines < maxlines)) { // end of Presentation Log header

				// check what type of event is in the line: Pulse, Picture or other
				if (readsubjectcolumn) {
					sscanf(str, "%s %d %s %s", Subject, &Trial, EventType, Code);
				} else {
					sscanf(str, "%d %s %s", &Trial, EventType, Code);
				}

				// if the type of event is 'Picture' or 'Sound', save the information
                                if (pictureEventsOnly) {
                                    if (strcmp(EventType, eventPicture) == 0) {

                                        if (readsubjectcolumn) {
                                                sscanf(str, "%s %d %s %s %lf %d %d %lf", Subject, &Trial, EventType, Code, &Time, &TTime, &Uncertainty, &Duration); // new in v05
                                        } else {
                                                sscanf(str, "%d %s %s %lf %d %d %lf", &Trial, EventType, Code, &Time, &TTime, &Uncertainty, &Duration);
                                        }
                                        char *code = new char[256];
                                        strcpy(code, Code);
                                        trials[currentcondition].code = code;
                                        trials[currentcondition].time = Time;
                                        trials[currentcondition].duration = Duration;

                                        sprintf(tmp3, "Trial: %d, code: %s, duration: %15.0lf", Trial, Code, (Duration/10));
                                        qxLogText(tmp3);
                                        currentcondition++;
                                    }
                                } else {
                                    if ((strcmp(EventType, eventPicture) == 0) || (strcmp(EventType, eventSound) == 0) ||
                                    (strcmp(EventType, eventVideo) == 0) || (strcmp(EventType, eventResponse) == 0)) {

					if (readsubjectcolumn) {
						sscanf(str, "%s %d %s %s %lf %d %d %lf", Subject, &Trial, EventType, Code, &Time, &TTime, &Uncertainty, &Duration); // new in v05
					} else {
						sscanf(str, "%d %s %s %lf %d %d %lf", &Trial, EventType, Code, &Time, &TTime, &Uncertainty, &Duration);
					}
					char *code = new char[256];
					strcpy(code, Code);
					trials[currentcondition].code = code;
					trials[currentcondition].time = Time;
					trials[currentcondition].duration = Duration;

					sprintf(tmp3, "Trial: %d, code: %s, duration: %15.0lf", Trial, Code, (Duration/10));
					qxLogText(tmp3);
					currentcondition++;
                                    }
                                } // end many types of Events
                                                        }
			if (nroflines == maxlines)	{
				endoffile = true;
			}

			strcpy(prevline, str);
			nroflines++;
		}

		pres->trials = trials;
		pres->nroftrials = nrofconditions;	// for now, the same
		pres->nrofcodes = nrofconditions;
		pres->origname = presFileName;
		pres->scenarioname = getFileName(presFileName, true, false);
		fclose(presFile);
	} catch (...) {
		qxLogText("Error while reading Presentation file");
	}
	return pres;
}

bool convertprotocol(presHdr *pres, int TR, int resolution, int margin, int durationtype) {

	bool success = true;
	char *txt = new char[256];

	// 1. Add the trials to BrainVoyager ProtocolCondition_Headers
	qxClearProtocol(); // bugfix: added in version 0.9.1
	Protocol_Header *prt = new Protocol_Header;
	qxLogText("Creating new protocol...");
	if (resolution == UNITS_VOLS) {
		prt->TimeResolution = UNITS_VOLS;
		strcpy(prt->ProtocolFileName, createFileName(exchangeExtension(pres->origname, ".prt"), "_VOLS"));
	} else {
		prt->TimeResolution = UNITS_MSEC;
		strcpy(prt->ProtocolFileName, createFileName(exchangeExtension(pres->origname, ".prt"), "_MSEC"));
	}
	sprintf(txt, "Creating protocol name:\n\n\t%s\n", prt->ProtocolFileName);
	qxLogText(txt);

	/* 2.	Loop through the conditions in the Presentation structure and convert to BrainVoyager
	 ProtocolCondition_Headers. Sort the trials into conditions and intervals
	 and add to current protocol in BrainVoyager
	 */
	success = sortConditionsAndIntervals(pres, prt, TR, resolution, durationtype);

	// 3. Add other stimulation protocol condition properties
	if (success) {
	//	qxLogText("Adding experiment name...");
	//	strcpy(prt->ExperimentName, pres->scenarioname);
		qxLogText("Setting colors and pen thickness...");
	//	setProtocolProperties(prt); // changed 31-08-11
		setProtocolProperties(prt, pres->scenarioname);
		strcpy(prt->ProtocolFileName, exchangeExtension(pres->origname, ".prt"));
		qxSetProtocol(prt);
		qxSaveProtocol(prt->ProtocolFileName);
	//	qxSetProtocol(prt); // reversed with Save at 31-08-11: experimentname and resolution is not saved properly
		qxLogText("Protocol generated from Presentation Log file.");
	} else {
		qxLogText("Creating protocol did not succeed.");
	}
	return success;
}

void setProtocolProperties(Protocol_Header *prt) {

	int plbgcolor = hexToDec("000000");
	int plrtcbgcolor = 1056964608 + 2000;
	int pltcolor = 1056964608 + 3000;
	int pltccolor = 1056964608 + 4000;
	prt->PlotBackgroundColor = plbgcolor;
	prt->PlotReferenceTimeCourseColor = plrtcbgcolor;
	prt->PlotReferenceTimeCoursePenThickness = 2;
	prt->PlotTextColor = pltcolor;
	prt->PlotTimeCourseColor = pltccolor;
	prt->PlotTimeCoursePenThickness = 2;
	prt->TimeResolution = 2;
}

void setProtocolProperties(Protocol_Header *prt, char *experimentname) {

        int plbgcolor = hexToDec("000000");
        int plrtcbgcolor = 1056964608 + 2000;
        int pltcolor = hexToDec("FFFFFF");
        int pltccolor = hexToDec("FFFFFF");
        prt->PlotBackgroundColor = plbgcolor;
        prt->PlotReferenceTimeCourseColor = plrtcbgcolor;
        prt->PlotReferenceTimeCoursePenThickness = 2;
        prt->PlotTextColor = pltcolor;
        prt->PlotTimeCourseColor = pltccolor;
        prt->PlotTimeCoursePenThickness = 2;
        char *expname = new char[301];
        sprintf(expname, experimentname);
        strcpy(prt->ExperimentName, expname);
}


bool sortConditionsAndIntervals(presHdr *conditions, Protocol_Header *prt, int TR, int resolution, int margin, int durationtype) {

	int condcounter = 0;
	bool success = true;
	int nroftrials = conditions->nroftrials;

	// 1. first create and prepare the list
	qxLogText("Creating and preparing conditions and intervals lists...");
	map<string, int> trialindex; // holds condition names and indices to Presentation structure
	map<string, int> conditioninterval; // holds condition names and number of intervals for condition
	map<string, int>::iterator iter;
	map<string, int>::iterator indexiterator;
	map<string, int> condmapping;

	// 2. while iterating, compare the names
	int trialcounter, keyisdefined;
	char *logtxt = new char[1024];
	condcounter = 0; // set to 0 again
	int startcolorval = 1056964608;

	// 2a. find unique condition names and add to map, add conditions
	qxLogText("Searching condition names...");
	for (trialcounter = 0; trialcounter < nroftrials; trialcounter++) {

		keyisdefined = condmapping.count(conditions->trials[trialcounter].code);
		if (keyisdefined == 0) {
			ProtocolCondition_Header *condition = new ProtocolCondition_Header;
			strcpy(condition->ConditionName, conditions->trials[trialcounter].code);
			condition->Color = startcolorval += 50005;
			qxAddProtocolCondition(condition);
			conditioninterval[conditions->trials[trialcounter].code] = 1;
			condmapping[conditions->trials[trialcounter].code] = condcounter;
			condcounter++;
		}  else {

			conditioninterval[conditions->trials[trialcounter].code] += 1; // increase number of intervals for this condition
		}
	}
	prt->NrOfConditions = condcounter;

	// 2b. loop over map with unique conditions and create BrainVoyager conditions and intervals
	int beginint, endint, oldduration, newduration;
	char *durtxt = new char[512];
	for (trialcounter = 0; trialcounter < nroftrials; trialcounter++) {

		beginint = intervalbegin(resolution, conditions->trials[trialcounter].time, TR, margin);
		if (durationtype == DURATION_LOG_BASED) {
			endint = intervalend(resolution, beginint, conditions->trials[trialcounter].duration, TR, margin);
		} else {
			oldduration = (conditions->trials[trialcounter+1].time - conditions->trials[trialcounter].time);
			newduration = (oldduration - 20);
			if (trialcounter < (nroftrials-1)) {
				endint = intervalend(resolution, beginint, newduration, TR, margin);
			} else {
				qxLogText("Last interval(s), computation based on Duration in protocol.");
				endint = intervalend(resolution, beginint, conditions->trials[trialcounter].duration, TR, margin);
			}
		}
		qxAddProtocolConditionInterval(condmapping[conditions->trials[trialcounter].code], beginint, endint);
		sprintf(logtxt, "Added interval: %d - %d, for condition: %s", beginint, endint, conditions->trials[trialcounter].code);
		qxLogText(logtxt);
	}

	return success;
}

bool changeProtocolOffset(Protocol_Header *prt, int calculationtype, int offset) {

	bool success = true;
	char *logtxt = new char[512];
	qxGetProtocol(prt);

	int ccounter, intervcounter;
	sprintf(logtxt, "Changing protocol offset with %d...", offset);
	qxLogText(logtxt);
	for (ccounter = 0; ccounter < prt->NrOfConditions; ccounter++) {
		ProtocolCondition_Header conds;
		qxGetProtocolCondition(ccounter, &conds);
		sprintf(logtxt, "\nCurrent condition: %s", conds.ConditionName);
		qxLogText(logtxt);
		for (intervcounter = 0; intervcounter < conds.NrOfIntervals; intervcounter++) {
			sprintf(logtxt, "Previous interval: %d - %d", conds.IntervalFrom[intervcounter], conds.IntervalTo[intervcounter]);
			qxLogText(logtxt);
			if (calculationtype == CALC_SUBTRACT) {
				conds.IntervalFrom[intervcounter] -= offset;
				if (conds.IntervalFrom[intervcounter] < 0) conds.IntervalFrom[intervcounter] = 0;
				conds.IntervalTo[intervcounter]	-= offset;
			} else {
				conds.IntervalFrom[intervcounter] += offset;
				conds.IntervalTo[intervcounter]	+= offset;
			}
			sprintf(logtxt, "New interval: %d - %d", conds.IntervalFrom[intervcounter], conds.IntervalTo[intervcounter]);
			qxLogText(logtxt);
		}
	}

        char *infix = new char[50];
        sprintf(infix, "%s%i", ((calculationtype == CALC_SUBTRACT) ? "_min" : "_plus"), offset);
        char *newname = createFileName(prt->ProtocolFileName, infix);
        //qxSetProtocol(newprt);
        qxSaveProtocol(newname);
        qxLogText("Protocol saved with new name:");
        qxLogText(newname);
        //qxSaveProtocol(prt->ProtocolFileName);
	return success;
}

int intervalbegin(int resolution, double time, int TR, int margin) {

	int mod = 0;
	int from;
//	char *txt = new char[512];
	int begin = int(time / 10);	 // convert 10s of msec to msec
	if (resolution == UNITS_MSEC) {
		from = begin;
	} else {
		mod = (begin % TR);
		if (mod > margin) {
			from = int(ceilf(begin/TR));
		} else {
			from = int(floorf(begin/TR));
		}
	//	sprintf(txt, "Margin: %d, (intervalbegin mod TR): %d, upper value: %d, lower value: %d", margin, mod, int(ceilf(begin/TR)), int(floorf(begin/TR)));
	//	qxLogText(txt);
	}
	return from;
}

int intervalend(int resolution, int start, double duration, int TR, int margin) {

	int mod = 0;
	int to;
	int initialto = start + int(floor((duration/10)));
	if (resolution == UNITS_MSEC) {
		to = initialto;
	} else {
		mod = (initialto % TR);
		if (mod > margin) {
			to = int(ceilf(initialto/TR));
		} else {
			to = int(floorf(initialto/TR));
		}
	}
	return to;
}

//char *getNewOrderOfConditions(Protocol_Header *prt) {
int getNewOrderOfConditions(Protocol_Header *prt, char *return_txt) {

	map<int, int> conditionsindex;
	string conditionlist = "Please enter the new order for the conditions, separated by comma's:\n";
	int conditioncounter;
	for (conditioncounter = 0; conditioncounter < prt->NrOfConditions; conditioncounter++) {

		ProtocolCondition_Header conds;
		ProtocolCondition_Header *currentcondition = new ProtocolCondition_Header;
		qxGetProtocolCondition(conditioncounter, currentcondition);
		char *loc = new char[256];
		sprintf(loc, "%d = %s\n", conditioncounter, currentcondition->ConditionName);
		conditionlist.append(loc);

	}
	conditionlist.append("\nExample: 3,4,5,1,6,2,7\n");
//	char *return_txt = new char[256];
	//qxGetTextInput(conditionlist.c_str(), "1,2,3,4", return_txt);
	int success = qxGetTextInput(conditionlist.c_str(), "1,2,3,4", return_txt);
//	return return_txt;
	return success;
}

bool convertConditionListToConditionVector(char *conditionlist, vector<int> &conditionvector) {

        bool success = true;
        int oldpos = 0;
        char * pch;
                pch = strtok (conditionlist,",");
                conditionvector[oldpos] = atoi(pch); // get first string part
                oldpos++;
                while (pch != NULL)	{
                        pch = strtok (NULL, ","); // get rest of string parts
                        if (pch != NULL) {
                                conditionvector[oldpos] = atoi(pch);
                        }
                        oldpos++;
                }

        return success;
}

bool makeAndSaveReorderedProtocol(Protocol_Header *origprt, vector<int> &conditionvector) {

        int newpos;
        bool success = true;
        int icounter;
        protocolstructure *prtcollection = new protocolstructure[origprt->NrOfConditions];
        char *txt = new char[256];

        // copy current protocol in protocolstructure
        for (newpos = 0; newpos < conditionvector.size(); newpos++) {

                ProtocolCondition_Header *currentcondition = new ProtocolCondition_Header;
                qxGetProtocolCondition(conditionvector[newpos], currentcondition);
                prtcollection[newpos].header = currentcondition;

                // let user know what happens
                sprintf(txt, "Condition name: %s\nNumber of intervals: %i", currentcondition->ConditionName, currentcondition->NrOfIntervals);
                qxLogText(txt);
                sprintf(txt, "New position: %i, previously: %i", newpos, conditionvector[newpos]);
                qxLogText(txt);

                // make deep copy of intervals, just to be sure
                int  *intervalFrom = new int[currentcondition->NrOfIntervals];
                int  *intervalTo = new int[currentcondition->NrOfIntervals];
                for (icounter = 0; icounter < currentcondition->NrOfIntervals; icounter++) {
                        intervalFrom[icounter] = currentcondition->IntervalFrom[icounter];
                        intervalTo[icounter] = currentcondition->IntervalTo[icounter];
                }
                prtcollection[newpos].interval_begin_list = intervalFrom;
                prtcollection[newpos].interval_end_list = intervalTo;
        }

        qxClearProtocol();
        Protocol_Header *newprt = new Protocol_Header;
        setProtocolProperties(newprt, origprt->ExperimentName);
        int counter, ccounter;
        for (counter = 0; counter < origprt->NrOfConditions; counter++) {

                ProtocolCondition_Header *currentcondition = prtcollection[counter].header;
                qxAddProtocolCondition(currentcondition);
                for (ccounter = 0; ccounter < currentcondition->NrOfIntervals; ccounter++) {

                                qxAddProtocolConditionInterval(counter, prtcollection[counter].interval_begin_list[ccounter],
                                        prtcollection[counter].interval_end_list[ccounter]);

                }
        }
        newprt->NrOfConditions = origprt->NrOfConditions;
        char *newname = createFileName(origprt->ProtocolFileName, "_ordered");
        qxSetProtocol(newprt);
        qxSaveProtocol(newname);
        qxLogText("Protocol saved with new name:");
        qxLogText(newname);
        return success;
}

bool makeAndSaveReorderedAndColoredProtocol(Protocol_Header *origprt, vector<int> &conditionvector) {

        int newpos;
        bool success = true;
        int icounter;
        protocolstructure *prtcollection = new protocolstructure[origprt->NrOfConditions];
        char *txt = new char[256];

        // copy current protocol in protocolstructure
        for (newpos = 0; newpos < conditionvector.size(); newpos++) {

                ProtocolCondition_Header *currentcondition = new ProtocolCondition_Header;
                qxGetProtocolCondition(conditionvector[newpos], currentcondition);
                prtcollection[newpos].header = currentcondition;

                // let user know what happens
                sprintf(txt, "Condition name: %s\nNumber of intervals: %i", currentcondition->ConditionName, currentcondition->NrOfIntervals);
                qxLogText(txt);
                sprintf(txt, "New position: %i, previously: %i", newpos, conditionvector[newpos]);
                qxLogText(txt);

                // make deep copy of intervals, just to be sure
                int  *intervalFrom = new int[currentcondition->NrOfIntervals];
                int  *intervalTo = new int[currentcondition->NrOfIntervals];
                for (icounter = 0; icounter < currentcondition->NrOfIntervals; icounter++) {
                        intervalFrom[icounter] = currentcondition->IntervalFrom[icounter];
                        intervalTo[icounter] = currentcondition->IntervalTo[icounter];
                }
                prtcollection[newpos].interval_begin_list = intervalFrom;
                prtcollection[newpos].interval_end_list = intervalTo;
        }

        qxClearProtocol();
        Protocol_Header *newprt = new Protocol_Header;
        setProtocolProperties(newprt, origprt->ExperimentName);

        // create protocol conditions
        int counter, ccounter;
        for (counter = 0; counter < origprt->NrOfConditions; counter++) {

                ProtocolCondition_Header *currentcondition = prtcollection[counter].header;
                bool is5x2 = (conditionvector.size() == 18);
                currentcondition->Color = STARTCOLOR + getAutoColor(counter, is5x2);
                qxAddProtocolCondition(currentcondition);
                for (ccounter = 0; ccounter < currentcondition->NrOfIntervals; ccounter++) {

                                qxAddProtocolConditionInterval(counter, prtcollection[counter].interval_begin_list[ccounter],
                                        prtcollection[counter].interval_end_list[ccounter]);

                }
        }
        newprt->NrOfConditions = origprt->NrOfConditions;
        char *newname = createFileName(origprt->ProtocolFileName, "_ordered_colored");
        qxSetProtocol(newprt);
        qxSaveProtocol(newname);
        qxLogText("Protocol saved with new name:");
        qxLogText(newname);
        return success;
}

bool makeAndSaveReorderedAndColoredProtocol(Protocol_Header *origprt, vector<int> &conditionvector, int typeOfAnalysis) {

        int newpos;
        bool success = true;
        int icounter;
        protocolstructure *prtcollection = new protocolstructure[origprt->NrOfConditions];
        char *txt = new char[256];

        // copy current protocol in protocolstructure
        for (newpos = 0; newpos < conditionvector.size(); newpos++) {

                ProtocolCondition_Header *currentcondition = new ProtocolCondition_Header;
                qxGetProtocolCondition(conditionvector[newpos], currentcondition);
                prtcollection[newpos].header = currentcondition;

                // let user know what happens
                sprintf(txt, "Condition name: %s\nNumber of intervals: %i", currentcondition->ConditionName, currentcondition->NrOfIntervals);
                qxLogText(txt);
                sprintf(txt, "New position: %i, previously: %i", newpos, conditionvector[newpos]);
                qxLogText(txt);

                // make deep copy of intervals, just to be sure
                int  *intervalFrom = new int[currentcondition->NrOfIntervals];
                int  *intervalTo = new int[currentcondition->NrOfIntervals];
                for (icounter = 0; icounter < currentcondition->NrOfIntervals; icounter++) {
                        intervalFrom[icounter] = currentcondition->IntervalFrom[icounter];
                        intervalTo[icounter] = currentcondition->IntervalTo[icounter];
                }
                prtcollection[newpos].interval_begin_list = intervalFrom;
                prtcollection[newpos].interval_end_list = intervalTo;
        }

        qxClearProtocol();
        Protocol_Header *newprt = new Protocol_Header;
        setProtocolProperties(newprt, origprt->ExperimentName);

        // create protocol conditions
        int counter, ccounter;
        for (counter = 0; counter < origprt->NrOfConditions; counter++) {

                ProtocolCondition_Header *currentcondition = prtcollection[counter].header;
                currentcondition->Color = STARTCOLOR + getMultiAutoColor(counter, typeOfAnalysis); // new per 12-10-09
                qxAddProtocolCondition(currentcondition);
                for (ccounter = 0; ccounter < currentcondition->NrOfIntervals; ccounter++) {

                     qxAddProtocolConditionInterval(counter, prtcollection[counter].interval_begin_list[ccounter],
                     prtcollection[counter].interval_end_list[ccounter]);

                }
        }
        newprt->NrOfConditions = origprt->NrOfConditions;
        char *newname = createFileName(origprt->ProtocolFileName, "_ordered_colored");
        qxSetProtocol(newprt);
        qxSaveProtocol(newname);
        qxLogText("Protocol saved with new name:");
        qxLogText(newname);
        return success;
}

